@@ -30,7 +30,9 @@ import shutil
 import six
 import tempfile
 
-__version__ = '2.0.0'
+from requests_toolbelt import MultipartEncoder
+
+__version__ = '2.1.0'
 __license__ = 'Apache 2.0'
 
 DEFAULT_PAGE_LIMIT = 50  # Number of results to fetch per request
@@ -88,6 +90,47 @@ class HttpError(Exception):
         return super(HttpError, self).__str__() + '\nResponse text: ' + self.responseText
 
 
+class _NoopProgressReporter(object):
+    reportProgress = False
+
+    def __init__(self, label='', length=0):
+        self.label = label
+        self.length = length
+
+    def update(self, chunkSize):
+        pass
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, tb):
+        pass
+
+
+# Used for fast non-multipart upload
+class _ProgressBytesIO(six.BytesIO):
+    def __init__(self, *args, **kwargs):
+        self.reporter = kwargs.pop('reporter')
+        six.BytesIO.__init__(self, *args, **kwargs)
+
+    def read(self, _size):
+        _chunk = six.BytesIO.read(self, _size)
+        self.reporter.update(len(_chunk))
+        return _chunk
+
+
+# Used for deprecated multipart upload
+class _ProgressMultiPartEncoder(MultipartEncoder):
+    def __init__(self, *args, **kwargs):
+        self.reporter = kwargs.pop('reporter')
+        MultipartEncoder.__init__(self, *args, **kwargs)
+
+    def read(self, _size):
+        _chunk = MultipartEncoder.read(self, _size)
+        self.reporter.update(len(_chunk))
+        return _chunk
+
+
 class GirderClient(object):
     """
     A class for interacting with the Girder RESTful API.
@@ -123,8 +166,13 @@ class GirderClient(object):
     # The current maximum chunk size for uploading file chunks
     MAX_CHUNK_SIZE = 1024 * 1024 * 64
 
+    DEFAULT_API_ROOT = 'api/v1'
+    DEFAULT_HOST = 'localhost'
+    DEFAULT_PORT = 443
+    DEFAULT_SCHEME = 'https'
+
     def __init__(self, host=None, port=None, apiRoot=None, scheme=None, apiUrl=None,
-                 cacheSettings=None):
+                 cacheSettings=None, progressReporterCls=None):
         """
         Construct a new GirderClient object, given a host name and port number,
         as well as a username and password which will be used in all requests
@@ -146,14 +194,24 @@ class GirderClient(object):
             to pass 443 for the port
         :param cacheSettings: Settings to use with the diskcache library, or
             None to disable caching.
+        :param progressReporterCls: the progress reporter class to instantiate. This class
+            is expected to be a context manager with a constructor accepting `label` and
+            `length` keyword arguments, an `update` method accepting a `chunkSize` argument and
+            a class attribute `reportProgress` set to True (It can conveniently be
+            initialized using `sys.stdout.isatty()`).
+            This defaults to :class:`_NoopProgressReporter`.
         """
         if apiUrl is None:
-            if apiRoot is None:
-                apiRoot = '/api/v1'
+            if not apiRoot:
+                apiRoot = self.DEFAULT_API_ROOT
+            # If needed, prepend '/'
+            if not apiRoot.startswith('/'):
+                apiRoot = '/' + apiRoot
 
-            self.scheme = scheme or 'http'
-            self.host = host or 'localhost'
-            self.port = port or (443 if scheme == 'https' else 80)
+            self.scheme = scheme or self.DEFAULT_SCHEME
+            self.host = host or self.DEFAULT_HOST
+            self.port = port or (
+                self.DEFAULT_PORT if self.scheme == 'https' else 80)
 
             self.urlBase = '%s://%s:%s%s' % (
                 self.scheme, self.host, str(self.port), apiRoot)
@@ -166,6 +224,7 @@ class GirderClient(object):
         self.token = ''
         self._folderUploadCallbacks = []
         self._itemUploadCallbacks = []
+        self._serverApiDescription = {}
         self.incomingMetadata = {}
         self.localMetadata = {}
 
@@ -174,6 +233,11 @@ class GirderClient(object):
         else:
             self.cache = diskcache.Cache(**cacheSettings)
 
+        if progressReporterCls is None:
+            progressReporterCls = _NoopProgressReporter
+
+        self.progressReporterCls = progressReporterCls
+
     def authenticate(self, username=None, password=None, interactive=False, apiKey=None):
         """
         Authenticate to Girder, storing the token that comes back to be used in
@@ -231,7 +295,68 @@ class GirderClient(object):
 
             self.token = resp['authToken']['token']
 
-    def sendRestRequest(self, method, path, parameters=None, data=None, files=None, json=None):
+    def getServerVersion(self, useCached=True):
+        """
+        Fetch server API version. By default, caches the version
+        such that future calls to this function do not make another request to
+        the server.
+
+        :param useCached: Whether to return the previously fetched value. Set
+            to False to force a re-fetch of the version from the server.
+        :type useCached: bool
+        :return: The API version as a list (e.g. ``['1', '0', '0']``)
+        """
+        return self.getServerAPIDescription(useCached)["info"]["version"].split('.')
+
+    def getServerAPIDescription(self, useCached=True):
+        """
+        Fetch server RESTful API description.
+
+        :param useCached: Whether to return the previously fetched value. Set
+            to False to force a re-fetch of the description from the server.
+        :type useCached: bool
+        :return: The API descriptions as a dict.
+
+        For example: ::
+
+            {
+                "basePath": "/api/v1",
+                "definitions": {},
+                "host": "girder.example.com",
+                "info": {
+                    "title": "Girder REST API",
+                    "version": "X.Y.Z"
+                },
+                "paths": {
+                    "/api_key": {
+                        "get": {
+                            "description": "Only site administrators [...]",
+                            "operationId": "api_key_listKeys",
+                            "parameters": [
+                                {
+                                    "description": "ID of the user whose keys to list.",
+                                    "in": "query",
+                                    "name": "userId",
+                                    "required": false,
+                                    "type": "string"
+                                },
+                                ...
+                            ]
+                        }.
+                        ...
+                    }
+                ...
+                }
+            }
+
+        """
+        if not self._serverApiDescription or not useCached:
+            self._serverApiDescription = self.get('describe')
+
+        return self._serverApiDescription
+
+    def sendRestRequest(self, method, path, parameters=None,
+                        data=None, files=None, json=None, headers=None):
         """
         This method looks up the appropriate method, constructs a request URL
         from the base URL, path, and parameters, and then sends the request. If
@@ -251,10 +376,12 @@ class GirderClient(object):
             as the key/value pairs in the request parameters.
         :type parameters: dict
         :param data: A dictionary, bytes or file-like object to send in the body.
-        :param files: A dictonary of 'name' => file-like-objects for multipart encoding upload.
+        :param files: A dictionary of 'name' => file-like-objects for multipart encoding upload.
         :type files: dict
         :param json: A JSON object to send in the request body.
         :type json: dict
+        :param headers: If present, a dictionary of headers to encode in the request.
+        :type headers: dict
         """
         if not parameters:
             parameters = {}
@@ -266,9 +393,13 @@ class GirderClient(object):
         url = self.urlBase + path
 
         # Make the request, passing parameters and authentication info
+        _headers = {'Girder-Token': self.token}
+        if isinstance(headers, dict):
+            _headers.update(headers)
+
         result = f(
             url, params=parameters, data=data, files=files, json=json,
-            headers={'Girder-Token': self.token})
+            headers=_headers)
 
         # If success, return the json object. Otherwise throw an exception.
         if result.status_code in (200, 201):
@@ -284,12 +415,12 @@ class GirderClient(object):
         """
         return self.sendRestRequest('GET', path, parameters)
 
-    def post(self, path, parameters=None, files=None, data=None, json=None):
+    def post(self, path, parameters=None, files=None, data=None, json=None, headers=None):
         """
         Convenience method to call :py:func:`sendRestRequest` with the 'POST' HTTP method.
         """
         return self.sendRestRequest('POST', path, parameters, files=files,
-                                    data=data, json=json)
+                                    data=data, json=json, headers=headers)
 
     def put(self, path, parameters=None, data=None, json=None):
         """
@@ -573,24 +704,6 @@ class GirderClient(object):
         }
         return self.put(path, params)
 
-    def _fileChunker(self, filepath, filesize=None):
-        """
-        Generator returning chunks of a file in MAX_CHUNK_SIZE increments.
-
-        :param filepath: path to file on disk.
-        :param filesize: size of file on disk if known.
-        """
-        if filesize is None:
-            filesize = os.path.getsize(filepath)
-        startbyte = 0
-        nextChunkSize = min(self.MAX_CHUNK_SIZE, filesize - startbyte)
-        with open(filepath, 'rb') as fd:
-            while nextChunkSize > 0:
-                chunk = fd.read(nextChunkSize)
-                yield (chunk, startbyte)
-                startbyte = startbyte + nextChunkSize
-                nextChunkSize = min(self.MAX_CHUNK_SIZE, filesize - startbyte)
-
     def isFileCurrent(self, itemId, filename, filepath):
         """
         Tests whether the passed in filepath exists in the item with itemId,
@@ -615,7 +728,8 @@ class GirderClient(object):
         # to upload anyway in this case also.
         return (None, False)
 
-    def uploadFileToItem(self, itemId, filepath, reference=None, mimeType=None):
+    def uploadFileToItem(self, itemId, filepath, reference=None, mimeType=None, filename=None,
+                         progressCallback=None):
         """
         Uploads a file to an item, in chunks.
         If ((the file already exists in the item with the same name and size)
@@ -627,9 +741,16 @@ class GirderClient(object):
         :type reference: str
         :param mimeType: MIME type for the file. Will be guessed if not passed.
         :type mimeType: str or None
+        :param filename: path with filename used in Girder. Defaults to basename of filepath.
+        :param progressCallback: If passed, will be called after each chunk
+            with progress information. It passes a single positional argument
+            to the callable which is a dict of information about progress.
+        :type progressCallback: callable
         :returns: the file that was created.
         """
-        filename = os.path.basename(filepath)
+        if filename is None:
+            filename = filepath
+        filename = os.path.basename(filename)
         filepath = os.path.abspath(filepath)
         filesize = os.path.getsize(filepath)
 
@@ -651,9 +772,7 @@ class GirderClient(object):
             if reference:
                 params['reference'] = reference
             obj = self.put(path, params)
-            if '_id' in obj:
-                uploadId = obj['_id']
-            else:
+            if '_id' not in obj:
                 raise Exception(
                     'After creating an upload token for replacing file '
                     'contents, expected an object with an id. Got instead: ' +
@@ -673,29 +792,13 @@ class GirderClient(object):
             if reference:
                 params['reference'] = reference
             obj = self.post('file', params)
-            if '_id' in obj:
-                uploadId = obj['_id']
-            else:
+            if '_id' not in obj:
                 raise Exception(
                     'After creating an upload token for a new file, expected '
                     'an object with an id. Got instead: ' + json.dumps(obj))
 
-        for chunk, startbyte in self._fileChunker(filepath, filesize):
-            parameters = {
-                'offset': startbyte,
-                'uploadId': uploadId
-            }
-            filedata = {
-                'chunk': chunk
-            }
-            path = 'file/chunk'
-            obj = self.post(path, parameters=parameters, files=filedata)
-
-            if '_id' not in obj:
-                raise Exception(
-                    'After uploading a file chunk, did not receive object with _id. Got instead: ' +
-                    json.dumps(obj))
-        return obj
+        with open(filepath, 'rb') as f:
+            return self._uploadContents(obj, f, filesize, progressCallback=progressCallback)
 
     def _uploadContents(self, uploadObj, stream, size, progressCallback=None):
         """
@@ -716,32 +819,49 @@ class GirderClient(object):
         """
         offset = 0
         uploadId = uploadObj['_id']
-        while True:
-            data = stream.read(min(self.MAX_CHUNK_SIZE, (size - offset)))
 
-            if not data:
-                break
+        with self.progressReporterCls(label=uploadObj['name'], length=size) as reporter:
 
-            params = {
-                'offset': offset,
-                'uploadId': uploadId
-            }
-            files = {
-                'chunk': data
-            }
-            uploadObj = self.post('file/chunk', parameters=params, files=files)
-            offset += len(data)
+            while True:
+                chunk = stream.read(min(self.MAX_CHUNK_SIZE, (size - offset)))
 
-            if '_id' not in uploadObj:
-                raise Exception(
-                    'After uploading a file chunk, did not receive object with _id. Got instead: ' +
-                    json.dumps(uploadObj))
+                if not chunk:
+                    break
 
-            if callable(progressCallback):
-                progressCallback({
-                    'current': offset,
-                    'total': size
-                })
+                if isinstance(chunk, six.text_type):
+                    chunk = chunk.encode('utf8')
+
+                if self.getServerVersion() >= ['2', '2']:
+                    uploadObj = self.post(
+                        'file/chunk?offset=%d&uploadId=%s' % (offset, uploadId),
+                        data=_ProgressBytesIO(chunk, reporter=reporter))
+                else:
+                    # Prior to version 2.2 the server only supported multipart uploads
+                    parameters = {
+                        'offset': offset,
+                        'uploadId': uploadId
+                    }
+
+                    m = _ProgressMultiPartEncoder(
+                        reporter=reporter,
+                        fields={'chunk': ('chunk', chunk, 'application/octet-stream')},
+                    )
+
+                    uploadObj = self.post('file/chunk', parameters=parameters,
+                                          data=m, headers={'Content-Type': m.content_type})
+
+                if '_id' not in uploadObj:
+                    raise Exception(
+                        'After uploading a file chunk, did not receive object with _id. '
+                        'Got instead: ' + json.dumps(uploadObj))
+
+                offset += len(chunk)
+
+                if callable(progressCallback):
+                    progressCallback({
+                        'current': offset,
+                        'total': size
+                    })
 
         if offset != size:
             self.delete('file/upload/' + uploadId)
@@ -895,12 +1015,19 @@ class GirderClient(object):
                 return
 
         # download to a tempfile
+        progressFileName = fileId
+        if isinstance(path, six.string_types):
+            progressFileName = os.path.basename(path)
         req = requests.get(
             '%sfile/%s/download' % (self.urlBase, fileId),
             stream=True, headers={'Girder-Token': self.token})
         with tempfile.NamedTemporaryFile(delete=False) as tmp:
-            for chunk in req.iter_content(chunk_size=REQ_BUFFER_SIZE):
-                tmp.write(chunk)
+            with self.progressReporterCls(
+                    label=progressFileName,
+                    length=int(req.headers['content-length'])) as reporter:
+                for chunk in req.iter_content(chunk_size=REQ_BUFFER_SIZE):
+                    reporter.update(len(chunk))
+                    tmp.write(chunk)
 
         # save file in cache
         if self.cache is not None:
@@ -1199,7 +1326,7 @@ class GirderClient(object):
         :param parentItemId: id of parent item in Girder to add file to
         :param filePath: full path to the file
         """
-        self.uploadFileToItem(parentItemId, filePath)
+        self.uploadFileToItem(parentItemId, filePath, filename=localFile)
 
     def _uploadAsItem(self, localFile, parentFolderId, filePath, reuseExisting=False, dryRun=False):
         """Function for doing an upload of a file as an item.
@@ -1209,7 +1336,8 @@ class GirderClient(object):
         :param reuseExisting: boolean indicating whether to accept an existing item
             of the same name in the same location, or create a new one instead
         """
-        print('Uploading Item from %s' % localFile)
+        if not self.progressReporterCls.reportProgress:
+            print('Uploading Item from %s' % localFile)
         if not dryRun:
             currentItem = self.loadOrCreateItem(
                 os.path.basename(localFile), parentFolderId, reuseExisting)
